@@ -58,6 +58,8 @@ import threading
 import keyword
 import inspect
 import contextlib
+import types
+from weakref import WeakKeyDictionary
 
 try:
     import typing
@@ -66,7 +68,7 @@ except ImportError:
 
 try:
     from traceback import StackSummary, FrameSummary
-except ImportError:
+except ImportError:  # StackSummary, FrameSummary were added in Python 3.5
 
     class _Dummy:
         pass
@@ -95,6 +97,14 @@ except NameError:  # Python3
 PY3 = sys.version_info[0] >= 3
 
 
+cfg_print_builtins = False
+cfg_print_not_found = False
+cfg_print_bound_methods = False
+cfg_print_modules = False
+cfg_print_module_functions = False
+cfg_print_module_classes = False
+
+
 def parse_py_statement(line):
     """
     Parse Python statement into tokens.
@@ -110,6 +120,10 @@ def parse_py_statement(line):
     """
     state = 0
     cur_token = ""
+    str_prefix = None
+    str_is_f_string = False  # whether we are in an f-string
+    str_quote = None
+    f_str_expr_opening_brackets = 0
     spaces = " \t\n"
     ops = ".,;:+-*/%&!=|(){}[]^<>"
     i = 0
@@ -132,29 +146,36 @@ def parse_py_statement(line):
                 yield "op", c
             elif c == "#":
                 state = 6
-            elif c == '"':
+            elif c in "\"'":
                 state = 1
-            elif c == "'":
-                state = 2
+                str_prefix = None
+                str_is_f_string = False
+                str_quote = c
+                cur_token = ""
             else:
                 cur_token = c
-                state = 3
-        elif state == 1:  # string via "
+                state = 3  # identifier
+        elif state == 1:  # string
             if c == "\\":
-                state = 4
-            elif c == '"':
-                yield "str", cur_token
-                cur_token = ""
-                state = 0
-            else:
-                cur_token += c
-        elif state == 2:  # string via '
-            if c == "\\":
-                state = 5
-            elif c == "'":
-                yield "str", cur_token
+                cur_token += _escape_char(line[i : i + 1])
+                i += 1
+            elif c == str_quote:
+                yield "str" if not str_prefix else "%s-str" % str_prefix, cur_token
                 cur_token = ""
                 state = 0
+            elif str_is_f_string and c == "{":  # f-string
+                if line[i - 1 : i + 1] == "{{":
+                    cur_token += "{"
+                    i += 1
+                else:
+                    yield "str" if not str_prefix else "%s-str" % str_prefix, cur_token
+                    yield "f-str-expr-open", "{"
+                    cur_token = ""
+                    f_str_expr_opening_brackets = 0
+                    state = 4
+            elif str_is_f_string and c == "}" and line[i - 1 : i + 1] == "}}":
+                cur_token += "}"
+                i += 1
             else:
                 cur_token += c
         elif state == 3:  # identifier
@@ -163,20 +184,40 @@ def parse_py_statement(line):
                 cur_token = ""
                 state = 0
                 i -= 1
-            elif c == '"':  # identifier is string prefix
-                cur_token = ""
+            elif c in "\"'":  # identifier is string prefix
                 state = 1
-            elif c == "'":  # identifier is string prefix
+                str_prefix = cur_token
+                str_is_f_string = "f" in str_prefix or "F" in str_prefix
+                str_quote = c
                 cur_token = ""
-                state = 2
             else:
                 cur_token += c
-        elif state == 4:  # escape in "
-            cur_token += _escape_char(c)
-            state = 1
-        elif state == 5:  # escape in '
-            cur_token += _escape_char(c)
-            state = 2
+        elif state == 4:  # f-string expression (like state 0 but simplified)
+            if c in spaces:
+                pass
+            elif c in ops:
+                if f_str_expr_opening_brackets == 0 and c == "}":
+                    yield "f-str-expr-close", "}"
+                    state = 1  # back into the f-string
+                    cur_token = ""
+                else:
+                    yield "op", c
+                    if c in "([{":
+                        f_str_expr_opening_brackets += 1
+                    elif c in ")]}":
+                        if f_str_expr_opening_brackets > 0:
+                            f_str_expr_opening_brackets -= 1
+            else:
+                cur_token = c
+                state = 5  # identifier in f-string expression
+        elif state == 5:  # identifier in f-string expression (like state 3 but simplified)
+            if c in spaces + ops:
+                yield "id", cur_token
+                cur_token = ""
+                state = 4
+                i -= 1
+            else:
+                cur_token += c
         elif state == 6:  # comment
             cur_token += c
     if state == 3:
@@ -233,18 +274,19 @@ def set_linecache(filename, source):
     """
     import linecache
 
+    # noinspection PyTypeChecker
     linecache.cache[filename] = None, None, [line + "\n" for line in source.splitlines()], filename
 
 
 # noinspection PyShadowingBuiltins
 def simple_debug_shell(globals, locals):
     """
-    :param dict[str] globals:
-    :param dict[str] locals:
+    :param dict[str,typing.Any] globals:
+    :param dict[str,typing.Any] locals:
     :return: nothing
     """
     try:
-        import readline
+        import readline  # noqa: F401
     except ImportError:
         pass  # ignore
     compile_string_fn = "<simple_debug_shell input>"
@@ -252,7 +294,7 @@ def simple_debug_shell(globals, locals):
         try:
             s = raw_input("> ")
         except (KeyboardInterrupt, EOFError):
-            print("breaked debug shell: " + sys.exc_info()[0].__name__)
+            print("broke debug shell: " + sys.exc_info()[0].__name__)
             break
         if s.strip() == "":
             continue
@@ -288,8 +330,8 @@ def debug_shell(user_ns, user_global_ns, traceback=None, execWrapper=None):
     Spawns some interactive shell. Tries to use IPython if available.
     Falls back to :func:`pdb.post_mortem` or :func:`simple_debug_shell`.
 
-    :param dict[str] user_ns:
-    :param dict[str] user_global_ns:
+    :param dict[str,typing.Any] user_ns:
+    :param dict[str,typing.Any] user_global_ns:
     :param traceback:
     :param execWrapper:
     :return: nothing
@@ -306,9 +348,6 @@ def debug_shell(user_ns, user_global_ns, traceback=None, execWrapper=None):
     if not ipshell and traceback and have_ipython:
         # noinspection PyBroadException
         try:
-            # noinspection PyPackageRequirements,PyUnresolvedReferences
-            from IPython.core.debugger import Pdb
-
             # noinspection PyPackageRequirements,PyUnresolvedReferences
             from IPython.terminal.debugger import TerminalPdb
 
@@ -460,7 +499,7 @@ def get_source_code(filename, lineno, module_globals=None):
     """
     :param str filename:
     :param int lineno:
-    :param dict[str]|None module_globals:
+    :param dict[str,typing.Any]|None module_globals:
     :return: source code of that line (including newline)
     :rtype: str
     """
@@ -693,7 +732,8 @@ class Color:
         ops = ".,;:+-*/%&!=|(){}[]^<>"
         i = 0
         cur_token = ""
-        color_args = {0: {}, len(s): {}}  # type: typing.Dict[int,typing.Dict[str]]  # pos in s -> color kwargs
+        # pos in s -> color kwargs
+        color_args = {0: {}, len(s): {}}  # type: typing.Dict[int,typing.Dict[str,typing.Any]]
 
         def finish_identifier():
             """
@@ -929,31 +969,36 @@ class _OutputLinesCollector:
         self.lines = []
         self.dom_term = DomTerm() if DomTerm.is_domterm() else None
 
-    def __call__(self, s1, s2=None, **kwargs):
+    def __call__(self, s1, s2=None, merge_into_prev=True, **kwargs):
         """
         Adds to self.lines.
         This strange function signature is for historical reasons.
 
         :param str s1:
         :param str|None s2:
+        :param bool merge_into_prev: if True and existing self.lines, merge into prev line.
         :param kwargs: passed to self.color
         """
         if kwargs:
             s1 = self.color(s1, **kwargs)
         if s2 is not None:
             s1 = add_indent_lines(s1, s2)
-        self.lines.append(s1 + "\n")
+        if merge_into_prev and self.lines:
+            self.lines[-1] += s1 + "\n"
+        else:
+            self.lines.append(s1 + "\n")
 
     @contextlib.contextmanager
-    def fold_text_ctx(self, line):
+    def fold_text_ctx(self, line, merge_into_prev=True):
         """
         Folds text, via :class:`DomTerm`, if available.
         Notes that this temporarily overwrites self.lines.
 
         :param str line: always visible
+        :param bool merge_into_prev: if True and existing self.lines, merge into prev line.
         """
         if not self.dom_term:
-            self.__call__(line)
+            self.__call__(line, merge_into_prev=merge_into_prev)
             yield
             return
         self.lines, old_lines = [], self.lines  # overwrite self.lines
@@ -969,7 +1014,10 @@ class _OutputLinesCollector:
             line = line[1:]
         self.dom_term.fold_text(line, hidden=hidden_text, file=output_buf, align=len(prefix))
         output_text = prefix[1:] + output_buf.getvalue()
-        self.lines.append(output_text)
+        if merge_into_prev and self.lines:
+            self.lines[-1] += output_text
+        else:
+            self.lines.append(output_text)
 
     def _pp_extra_info(self, obj, depth_limit=3):
         """
@@ -1008,10 +1056,15 @@ class _OutputLinesCollector:
 
     def pretty_print(self, obj):
         """
-        :param object obj:
+        :param typing.Any obj:
         :rtype: str
         """
-        s = repr(obj)
+        if isinstance(obj, types.FunctionType) and hasattr(obj, "__module__") and hasattr(obj, "__qualname__"):
+            s = "<function %s.%s>" % (obj.__module__, obj.__qualname__)
+        elif isinstance(obj, type) and hasattr(obj, "__module__") and hasattr(obj, "__qualname__"):
+            s = "<class %s.%s>" % (obj.__module__, obj.__qualname__)
+        else:
+            s = repr(obj)
         limit = output_limit()
         if len(s) > limit:
             if self.dom_term:
@@ -1031,16 +1084,35 @@ class _OutputLinesCollector:
 
 # For compatibility, we keep non-PEP8 argument names.
 # noinspection PyPep8Naming
-def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=False, with_color=None, with_vars=None):
+def format_tb(
+    tb=None,
+    limit=None,
+    allLocals=None,
+    allGlobals=None,
+    withTitle=False,
+    with_color=None,
+    with_vars=None,
+    clear_frames=True,
+):
     """
-    :param types.TracebackType|types.FrameType|StackSummary tb: traceback. if None, will use sys._getframe
+    Formats a traceback into a list of strings, each corresponding to one frame.
+
+    Replacement for traceback.format_tb.
+
+    :param types.TracebackType|types.FrameType|StackSummary tb: traceback. If None, will use sys._getframe
     :param int|None limit: limit the traceback to this number of frames. by default, will look at sys.tracebacklimit
-    :param dict[str]|None allLocals: if set, will update it with all locals from all frames
-    :param dict[str]|None allGlobals: if set, will update it with all globals from all frames
+    :param dict[str,typing.Any]|None allLocals: if set, will update it with all locals from all frames
+    :param dict[str,typing.Any]|None allGlobals: if set, will update it with all globals from all frames
     :param bool withTitle:
     :param bool|None with_color: output with ANSI escape codes for color
-    :param bool with_vars: will print var content which are referenced in the source code line. by default enabled.
-    :return: list of strings (line-based)
+    :param bool with_vars: will print var contents that are referenced in the source code line. by default enabled.
+    :param bool clear_frames: whether to call frame.clear() after processing it.
+        That will potentially fix some mem leaks regarding locals, so it can be important.
+        Also see https://github.com/python/cpython/issues/113939.
+        However, any further access to frame locals will not work (e.g., if you want to use a debugger afterward).
+    :return: list of strings, each corresponding to one frame in the traceback.
+        Each string contains the file name, line number, function name, source code line, maybe relevant variables,
+        etc., and a final newline.
     :rtype: list[str]
     """
     color = Color(enable=with_color)
@@ -1100,6 +1172,8 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
                 output("(Exclude vars because we are on a GC stack.)")
     if with_vars is None:
         with_vars = True
+    locals_start_str = color("    locals:", color.fg_colors[0])
+
     # noinspection PyBroadException
     try:
         if limit is None:
@@ -1108,49 +1182,15 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
         n = 0
         _tb = tb
 
-        class NotFound(Exception):
-            """
-            Identifier not found.
-            """
-
-        def _resolve_identifier(namespace, keys):
-            """
-            :param dict[str] namespace:
-            :param tuple[str] keys:
-            :return: namespace[name[0]][name[1]]...
-            """
-            if keys[0] not in namespace:
-                raise NotFound()
-            obj = namespace[keys[0]]
-            for part in keys[1:]:
-                obj = getattr(obj, part)
-            return obj
-
-        # noinspection PyShadowingNames
-        def _try_set(old, prefix, func):
-            """
-            :param None|str old:
-            :param str prefix:
-            :param func:
-            :return: old
-            """
-            if old is not None:
-                return old
-            try:
-                return add_indent_lines(prefix, func())
-            except NotFound:
-                return old
-            except Exception as e:
-                return prefix + "!" + e.__class__.__name__ + ": " + str(e)
-
         while _tb is not None and (limit is None or n < limit):
             if isframe(_tb):
                 f = _tb
             elif is_stack_summary(_tb):
-                if isinstance(_tb[0], ExtendedFrameSummary):
-                    f = _tb[0].tb_frame
+                _tb0 = _tb[0]
+                if isinstance(_tb0, ExtendedFrameSummary):
+                    f = _tb0.tb_frame
                 else:
-                    f = DummyFrame.from_frame_summary(_tb[0])
+                    f = DummyFrame.from_frame_summary(_tb0)
             else:
                 f = _tb.tb_frame
             if allLocals is not None:
@@ -1183,7 +1223,7 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
                     name,
                 ]
             )
-            with output.fold_text_ctx(file_descr):
+            with output.fold_text_ctx(file_descr, merge_into_prev=False):
                 source_code = get_source_code(filename, lineno, f.f_globals)
                 if source_code:
                     source_code = remove_indent_lines(replace_tab_indents(source_code)).rstrip()
@@ -1193,50 +1233,107 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
                     elif isinstance(f, DummyFrame) and not f.have_vars_available:
                         pass
                     else:
-                        with output.fold_text_ctx(color("    locals:", color.fg_colors[0])):
-                            already_printed_locals = set()  # type: typing.Set[typing.Tuple[str,...]]
+                        with output.fold_text_ctx(locals_start_str):
+                            already_covered_locals = set()  # type: typing.Set[typing.Tuple[str,...]]
+                            num_printed_locals = 0
                             for token_str in grep_full_py_identifiers(parse_py_statement(source_code)):
                                 splitted_token = tuple(token_str.split("."))
                                 for token in [splitted_token[0:i] for i in range(1, len(splitted_token) + 1)]:
-                                    if token in already_printed_locals:
+                                    if token in already_covered_locals:
                                         continue
-                                    token_value = None
-                                    token_value = _try_set(
-                                        token_value,
-                                        color("<local> ", color.fg_colors[0]),
-                                        lambda: format_py_obj(_resolve_identifier(f.f_locals, token)),
-                                    )
-                                    token_value = _try_set(
-                                        token_value,
-                                        color("<global> ", color.fg_colors[0]),
-                                        lambda: format_py_obj(_resolve_identifier(f.f_globals, token)),
-                                    )
-                                    token_value = _try_set(
-                                        token_value,
-                                        color("<builtin> ", color.fg_colors[0]),
-                                        lambda: format_py_obj(_resolve_identifier(f.f_builtins, token)),
-                                    )
-                                    token_value = token_value or color("<not found>", color.fg_colors[0])
+                                    already_covered_locals.add(token)
+                                    if token[0] in f.f_locals:
+                                        token_base_dict = f.f_locals
+                                        token_prefix_str = color("<local> ", color.fg_colors[0])
+                                    elif token[0] in f.f_globals:
+                                        token_base_dict = f.f_globals
+                                        token_prefix_str = color("<global> ", color.fg_colors[0])
+                                        if (
+                                            not cfg_print_module_functions
+                                            and len(token) == 1
+                                            and _is_module_function(token_base_dict, token[0], obj_is_dict=True)
+                                        ):
+                                            continue
+                                        if (
+                                            not cfg_print_module_classes
+                                            and len(token) == 1
+                                            and _is_module_class(token_base_dict, token[0], obj_is_dict=True)
+                                        ):
+                                            continue
+                                    elif token[0] in f.f_builtins:
+                                        if not cfg_print_builtins:
+                                            continue
+                                        token_base_dict = f.f_builtins
+                                        token_prefix_str = color("<builtin> ", color.fg_colors[0])
+                                    else:
+                                        if not cfg_print_not_found:
+                                            continue
+                                        token_base_dict = None
+                                        token_prefix_str = None
+
                                     prefix = "      %s " % color(".", color.fg_colors[0], bold=True).join(
                                         token
                                     ) + color("= ", color.fg_colors[0], bold=True)
-                                    output(prefix, token_value)
-                                    already_printed_locals.add(token)
-                            if len(already_printed_locals) == 0:
-                                output(color("       no locals", color.fg_colors[0]))
-                else:
+
+                                    if token_prefix_str is None:  # not found
+                                        token_repr = color("<not found>", color.fg_colors[0])
+                                    else:
+                                        try:
+                                            token_parent_obj = None
+                                            token_obj = token_base_dict[token[0]]
+                                            for attr in token[1:]:
+                                                token_parent_obj = token_obj
+                                                token_obj = getattr(token_obj, attr)
+                                        except Exception as e:
+                                            token_repr = token_prefix_str + "!" + e.__class__.__name__ + ": " + str(e)
+                                        else:  # found
+                                            if (
+                                                not cfg_print_bound_methods
+                                                and token_parent_obj is not None
+                                                and _is_bound_method(token_parent_obj, token[-1])
+                                            ):
+                                                continue
+                                            if not cfg_print_modules and isinstance(token_obj, types.ModuleType):
+                                                continue
+                                            if (
+                                                not cfg_print_module_functions
+                                                and token_parent_obj is not None
+                                                and _is_module_function(token_parent_obj, token[-1])
+                                            ):
+                                                continue
+                                            if (
+                                                not cfg_print_module_classes
+                                                and token_parent_obj is not None
+                                                and _is_module_class(token_parent_obj, token[-1])
+                                            ):
+                                                continue
+                                            token_repr = add_indent_lines(token_prefix_str, format_py_obj(token_obj))
+
+                                    output(prefix, token_repr)
+                                    num_printed_locals += 1
+
+                            if num_printed_locals == 0:
+                                if output.lines and output.lines[-1].endswith(locals_start_str + "\n"):
+                                    output.lines[-1] = output.lines[-1][: -len(locals_start_str) - 1]
+                                elif output.lines and output.lines[-1].endswith("\n"):
+                                    output.lines[-1] = output.lines[-1][:-1] + color(" none", color.fg_colors[0]) + "\n"
+                                else:
+                                    output(color("       no locals", color.fg_colors[0]))
+
+                else:  # no source code available
                     output(color("    -- code not available --", color.fg_colors[0]))
 
-            # Just like :func:`traceback.clear_frames`, but has an additional fix
-            # (https://github.com/python/cpython/issues/113939).
-            try:
-                f.clear()
-            except RuntimeError:
-                pass
-            else:
-                # Using this code triggers that the ref actually goes out of scope, otherwise it does not!
-                # https://github.com/python/cpython/issues/113939
-                f.f_locals  # noqa
+            if clear_frames:
+                # Just like :func:`traceback.clear_frames`, but has an additional fix
+                # (https://github.com/python/cpython/issues/113939).
+                try:
+                    f.clear()
+                except RuntimeError:
+                    pass
+                else:
+                    # Using this code triggers that the ref actually goes out of scope, otherwise it does not!
+                    # https://github.com/python/cpython/issues/113939
+                    f.f_locals  # noqa
 
             if isframe(_tb):
                 _tb = _tb.f_back
@@ -1260,6 +1357,8 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
 
 def print_tb(tb, file=None, **kwargs):
     """
+    Prints the traceback to stderr, or the given file.
+
     Replacement for traceback.print_tb.
 
     :param types.TracebackType|types.FrameType|StackSummary tb:
@@ -1329,6 +1428,13 @@ def better_exchook(
     if file is None:
         file = sys.stderr
 
+    if autodebugshell:
+        # noinspection PyBroadException
+        try:
+            debugshell = int(os.environ["DEBUG"]) != 0
+        except Exception:
+            pass
+
     color = Color(enable=with_color)
     output = _OutputLinesCollector(color=color)
 
@@ -1369,6 +1475,7 @@ def better_exchook(
                 allGlobals=all_globals,
                 withTitle=True,
                 with_color=color.enable,
+                clear_frames=not debugshell,
             )
         )
     else:
@@ -1451,14 +1558,9 @@ def better_exchook(
         file.write(line)
     file.flush()
 
-    if autodebugshell:
-        # noinspection PyBroadException
-        try:
-            debugshell = int(os.environ["DEBUG"]) != 0
-        except Exception:
-            pass
     if debugshell:
-        output("---------- DEBUG SHELL -----------")
+        file.write("---------- DEBUG SHELL -----------\n")
+        file.flush()
         debug_shell(user_ns=all_locals, user_global_ns=all_globals, traceback=tb)
 
 
@@ -1538,6 +1640,9 @@ def get_func_str_from_code_object(co, frame=None):
     return co.co_name
 
 
+_func_from_code_object_cache = WeakKeyDictionary()  # code object -> function
+
+
 def get_func_from_code_object(co, frame=None):
     """
     :param types.CodeType co:
@@ -1554,6 +1659,11 @@ def get_func_from_code_object(co, frame=None):
     import types
 
     assert isinstance(co, (types.CodeType, DummyFrame))
+    co_is_code_object = isinstance(co, types.CodeType)
+    if co_is_code_object:
+        candidate = _func_from_code_object_cache.get(co)
+        if candidate:
+            return candidate
     _attr_name = "__code__" if PY3 else "func_code"
     if frame and frame.f_code.co_nlocals > 0:
         func_name = frame.f_code.co_name
@@ -1561,18 +1671,23 @@ def get_func_from_code_object(co, frame=None):
         if frame_self is not None:
             candidate = getattr(frame_self.__class__, func_name, None)
             if candidate and (getattr(candidate, _attr_name, None) is co or isinstance(co, DummyFrame)):
+                if co_is_code_object:
+                    _func_from_code_object_cache[co] = candidate
                 return candidate
     try:
         candidate = getattr(_get_loaded_module_from_filename(co.co_filename), co.co_name, None)
     except ImportError:  # some modules have lazy loaders, but those might fail here
         candidate = None
     if candidate and (getattr(candidate, _attr_name, None) is co or isinstance(co, DummyFrame)):
+        if co_is_code_object:
+            _func_from_code_object_cache[co] = candidate
         return candidate
     if isinstance(co, DummyFrame):
         return None
     candidates = gc.get_referrers(co)
     candidates = [f for f in candidates if getattr(f, _attr_name, None) is co]
     if candidates:
+        _func_from_code_object_cache[co] = candidates[0]
         return candidates[0]
     return None
 
@@ -1645,10 +1760,11 @@ def iter_traceback(tb=None, enforce_most_recent_call_first=False):
         if is_frame(_tb):
             frame = _tb
         elif is_stack_summary(_tb):
-            if isinstance(_tb[0], ExtendedFrameSummary):
-                frame = _tb[0].tb_frame
+            _tb0 = _tb[0]
+            if isinstance(_tb0, ExtendedFrameSummary):
+                frame = _tb0.tb_frame
             else:
-                frame = DummyFrame.from_frame_summary(_tb[0])
+                frame = DummyFrame.from_frame_summary(_tb0)
         else:
             frame = _tb.tb_frame
         yield frame
@@ -1671,6 +1787,14 @@ class ExtendedFrameSummary(FrameSummary):
         super(ExtendedFrameSummary, self).__init__(**kwargs)
         self.tb_frame = frame
 
+    def __reduce__(self):
+        # We deliberately serialize this as just a FrameSummary,
+        # excluding the tb_frame, as this cannot be serialized (via pickle at least),
+        # and also we do not want this to be serialized.
+        # We also exclude _line and locals as it's not so easy to add them here,
+        # and also not so important.
+        return FrameSummary, (self.filename, self.lineno, self.name)
+
 
 class DummyFrame:
     """
@@ -1735,6 +1859,49 @@ def _StackSummary_extract(frame_gen, limit=None, lookup_lines=True, capture_loca
     return result
 
 
+def _is_bound_method(obj, attr_name):
+    if not PY3:
+        return False  # not properly supported in Python 2
+
+    meth = getattr(obj, attr_name, None)
+    meth = inspect.unwrap(meth)
+
+    if isinstance(meth, types.MethodType):
+        if meth.__self__ is not obj:
+            return False
+        cls = type(obj)
+        func = getattr(cls, attr_name, None)
+        return meth.__func__ is func
+
+    elif isinstance(meth, (types.BuiltinMethodType, getattr(types, "MethodWrapperType", types.BuiltinMethodType))):
+        if meth.__self__ is not obj:
+            return False
+        return meth.__name__ == attr_name
+
+    else:
+        return False
+
+
+def _is_module_function(obj, attr_name, obj_is_dict=False):
+    if obj_is_dict:
+        func = obj.get(attr_name, None)
+    else:
+        if not isinstance(obj, types.ModuleType):
+            return False
+        func = getattr(obj, attr_name, None)
+    return isinstance(func, types.FunctionType)
+
+
+def _is_module_class(obj, attr_name, obj_is_dict=False):
+    if obj_is_dict:
+        cls = obj.get(attr_name, None)
+    else:
+        if not isinstance(obj, types.ModuleType):
+            return False
+        cls = getattr(obj, attr_name, None)
+    return isinstance(cls, type)
+
+
 def install():
     """
     Replaces sys.excepthook by our better_exchook.
